/*
** Module   : PropertyHelper.java
** Abstract : Contains helper methods related to DMO properties
**
** Copyright (c) 2004-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ------------------Description-------------------
** 001 ECF 20051113   @23390 Created initial version. Contains helper
**                           methods related to DMO properties. Uses
**                           introspection to create maps of different
**                           property access categories: getters, setters,
**                           and validators.
** 002 ECF 20060504   @26000 Added sizersByProperty method. Ensure correct
**                           method variant is selected if methods are
**                           overloaded.
** 003 ECF 20060908   @29381 Fixed defect in makePropertyName(). The
**                           algorithm was unconditionally decapitalizing
**                           the property name's first letter, even when
**                           it should have remained uppercase.
** 004 ECF 20070531   @33908 Added methods to map methods to property
**                           names. Added caching of property/method maps
**                           to avoid redundant introspection.
** 005 ECF 20070531   @33913 Removed validatorsByProperty(). This method
**                           is no longer used and was performing
**                           unnecessary introspection.
** 006 ECF 20080521   @38500 Added settersByProperty(String) method
**                           variant. Integrated generics.
** 007 ECF 20080605   @38609 Made settersByProperty(String) method more
**                           lenient. It was throwing an
**                           IllegalStateException if invoked before
**                           settersByProperty(Class<?>).
** 008 ECF 20080925   @39947 Changed makePropertyName(). Property name is
**                           now intern()'d to save memory.
** 009 ECF 20090217   @41331 Added gettersForInterface() and backing map.
**                           Provides set of getter methods for a given
**                           DMO interface.
** 010 ECF 20090315   @41606 Added lazyExtentTriggersForInterface() and
**                           backing map. Provides list of getter methods
**                           which will trigger lazy initialization of the
**                           persistent collections used to manage extent
**                           fields in a DMO.
** 011 ECF 20090815   @43633 Added propertiesByGetter(), extentsByProperty().
**                           Needed by RecordBuffer.
** 012 ECF 20090817   @43681 Fixed gettersForInterface(). This method was not
**                           picking up the new getters for extent fields
**                           which accept NumberType for the index parameter.
** 013 VMN 20140227          Added support for custom denormalization of fields with extent and
**                           method getPropertyName().
*/

package com.goldencode.p2j.persist;

import java.beans.Introspector;
import java.lang.reflect.*;
import java.util.*;
import com.goldencode.p2j.util.*;

/**
 * Helper class which uses introspection to produce maps of different
 * categories of property-related methods:  getters, setters, validators, and
 * size methods (for retrieving the size of a list or composite property).
 * <p>
 * Maps are cached to avoid redundant introspection using reflection.
 */
public final class PropertyHelper
{
   /** Cache of property name to setter method maps, by DMO interface */
   private static Map<Class<?>, Map<String, Method>> p2sCache =
      new HashMap<Class<?>, Map<String, Method>>();
   
   /** Cache of property name to getter method maps, by DMO interface */
   private static Map<Class<?>, Map<String, Method>> p2gCache =
      new HashMap<Class<?>, Map<String, Method>>();
   
   /** Cache of property name to sizer method maps, by DMO interface */
   private static Map<Class<?>, Map<String, Method>> p2zCache =
      new HashMap<Class<?>, Map<String, Method>>();
   
   /** Cache of property name to extent maps, by DMO interface */
   private static Map<Class<?>, Map<String, Integer>> p2xCache =
      new HashMap<Class<?>, Map<String, Integer>>();
   
   /** Cache of getter method to property name maps, by DMO interface */
   private static Map<Class<?>, Map<Method, String>> g2pCache =
      new HashMap<Class<?>, Map<Method, String>>();
   
   /** Cache of setter method to property name maps, by DMO interface */
   private static Map<Class<?>, Map<Method, String>> s2pCache =
      new HashMap<Class<?>, Map<Method, String>>();
   
   /** Cache of property name to setter method maps, by DMO entity name */
   private static Map<String, Map<String, Method>> p2sEntityCache =
      new HashMap<String, Map<String, Method>>();
   
   /** Cache of getter method sets, by DMO interface */
   private static Map<Class<?>, Set<Method>> i2gCache =
      new HashMap<Class<?>, Set<Method>>();
   
   /** Cache of methods to trigger lazy collection init, by DMO interface */
   private static Map<Class<?>, List<Method>> i2ltCache =
      new HashMap<Class<?>, List<Method>>();
   
   /**
    * Private constructor;  all access is via static methods.
    */
   private PropertyHelper()
   {
   }
   
   /**
    * Retrieve a map of all property setter methods for the given DMO entity.
    * The keys are the properties' names;  the values are the methods.
    * <p>
    * Setter methods are all those declared by the corresponding interface,
    * whose names begin with <code>set</code>, excluding those which accept a
    * collection parameter as the value to set.
    *
    * @param   entityName
    *          DMO implementation class name.
    *
    * @return  Map of DMO property names to setter methods.
    */
   public static Map<String, Method> settersByProperty(String entityName)
   {
      Map<String, Method> map = null;
      
      synchronized (p2sEntityCache)
      {
         map = p2sEntityCache.get(entityName);
      }
      
      if (map == null)
      {
         Class<?> dmoClass = DBUtils.dmoClassForEntity(entityName);
         Class<?> dmoIface = DMOIndex.getDMOInterface(dmoClass);
         
         // This call will populate the p2sEntityCache for faster future
         // access.
         map = settersByProperty(dmoIface);
      }
      
      return map;
   }
   
   /**
    * Build a map of all property setter methods for the given DMO interface.
    * The keys are the properties' names;  the values are the methods.
    * <p>
    * Setter methods are all those declared by the interface, whose names
    * begin with <code>set</code>, excluding those which accept a collection
    * parameter as the value to set.
    *
    * @param   dmoIface
    *          DMO interface to introspect.
    *
    * @return  Map of DMO property names to setter methods.
    */
   public static Map<String, Method> settersByProperty(Class<?> dmoIface)
   {
      synchronized (p2sCache)
      {
         Map<String, Method> map = p2sCache.get(dmoIface);
         
         if (map == null)
         {
            // Map all setter methods.
            map = methodsByProperty(dmoIface.getDeclaredMethods(),
                                    new String[] { "set" });
            
            // Remove those that set collections (only Hibernate needs these).
            Iterator<Method> iter = map.values().iterator();
            while (iter.hasNext())
            {
               Method meth = iter.next();
               Class<?>[] parms = meth.getParameterTypes();
               int len = parms.length;
               if (len > 0 &&
                   Collection.class.isAssignableFrom(parms[len - 1]))
               {
                  iter.remove();
               }
            }
            
            map = Collections.unmodifiableMap(map);
            p2sCache.put(dmoIface, map);
            
            String entity = DBUtils.entityForDMOInterface(dmoIface);
            synchronized (p2sEntityCache)
            {
               p2sEntityCache.put(entity, map);
            }
         }
         
         return map;
      }
   }
   
   /**
    * Build a map of all property getter methods for the given DMO interface.
    * The keys are the properties' names;  the values are the methods.
    * <p>
    * Getter methods are all those declared by the interface, whose names
    * begin with <code>get</code> or <code>is</code>, excluding those which
    * return a collection.
    *
    * @param   dmoIface
    *          DMO interface to introspect.
    *
    * @return  Map of DMO property names to getter methods.
    */
   public static Map<String, Method> gettersByProperty(Class<?> dmoIface)
   {
      synchronized (p2gCache)
      {
         Map<String, Method> map = p2gCache.get(dmoIface);
         
         if (map == null)
         {
            // Map all getter methods.
            map = methodsByProperty(dmoIface.getDeclaredMethods(),
                                    new String[] { "get", "is" });
            
            // Remove those that return collections (only Hibernate needs
            // these).
            Iterator<Method> iter = map.values().iterator();
            while (iter.hasNext())
            {
               Method meth = iter.next();
               if (Collection.class.isAssignableFrom(meth.getReturnType()))
               {
                  iter.remove();
               }
            }
            
            map = Collections.unmodifiableMap(map);
            p2gCache.put(dmoIface, map);
         }
         
         return map;
      }
   }
   
   /**
    * Build a map of all list size query methods for the given DMO interface.
    * The keys are the properties' names;  the values are the methods.
    * <p>
    * Sizer methods are all those declared by the interface, whose names
    * begin with <code>size</code>.
    *
    * @param   dmoIface
    *          DMO interface to introspect.
    *
    * @return  Map of DMO property names to sizer methods.
    */
   static Map<String, Method> sizersByProperty(Class<?> dmoIface)
   {
      synchronized (p2zCache)
      {
         Map<String, Method> map = p2zCache.get(dmoIface);
         
         if (map == null)
         {
            map = methodsByProperty(dmoIface.getDeclaredMethods(),
                                    new String[] { "size" });
            
            map = Collections.unmodifiableMap(map);
            p2zCache.put(dmoIface, map);
         }
         
         return map;
      }
   }
   
   /**
    * Get a modifiable, synchronized map to be used to store and retrieve
    * extent values for a given DMO type.  Upon initial creation, it is not
    * pre-populated with these values, because that information is not
    * available to this class, but the contract of a caller to this method is
    * that if it finds an entry for a particular, indexed property is missing,
    * it should query the value and add it to the map.  The map is thread safe
    * and is cached.
    * 
    * @param   dmoIface
    *          DMO interface.
    * 
    * @return  A map as described above.
    */
   static Map<String, Integer> extentsByProperty(Class<?> dmoIface)
   {
      synchronized (p2xCache)
      {
         Map<String, Integer> map = p2xCache.get(dmoIface);
         
         if (map == null)
         {
            map = new HashMap<String, Integer>();
            map = Collections.synchronizedMap(map);
            p2xCache.put(dmoIface, map);
         }
         
         return map;
      }
   }
   
   /**
    * Build a map of all property names for the given DMO interface, where
    * the keys are methods which retrieve those properties.  This will include
    * simple as well as indexed getter methods.
    *
    * @param   dmoIface
    *          DMO interface to introspect.
    *
    * @return  Map of DMO getter methods to property names.
    */
   static Map<Method, String> propertiesByGetter(Class<?> dmoIface)
   {
      synchronized (g2pCache)
      {
         Map<Method, String> map = g2pCache.get(dmoIface);
         
         if (map == null)
         {
            // Map all properties by setter method.
            map = propertiesByMethod(dmoIface.getDeclaredMethods(),
                                     new String[] { "is", "get" });
            
            // Remove those that get collections (only Hibernate needs these).
            Iterator<Method> iter = map.keySet().iterator();
            while (iter.hasNext())
            {
               Method meth = iter.next();
               Class<?> ret = meth.getReturnType();
               if (Collection.class.isAssignableFrom(ret))
               {
                  iter.remove();
               }
            }
            
            map = Collections.unmodifiableMap(map);
            g2pCache.put(dmoIface, map);
         }
         
         return map;
      }
   }
   
   /**
    * Build a map of all property names for the given DMO interface, where
    * the keys are methods which updates those properties.  This will include
    * indexed setter methods, as well as bulk setter variants which accept
    * array and scalar values to be applied to multiple indices of a list
    * or composite property.
    *
    * @param   dmoIface
    *          DMO interface to introspect.
    *
    * @return  Map of DMO setter methods to property names.
    */
   static Map<Method, String> propertiesBySetter(Class<?> dmoIface)
   {
      synchronized (s2pCache)
      {
         Map<Method, String> map = s2pCache.get(dmoIface);
         
         if (map == null)
         {
            // Map all properties by setter method.
            map = propertiesByMethod(dmoIface.getDeclaredMethods(),
                                     new String[] { "set" });
            
            // Remove those that set collections (only Hibernate needs these).
            Iterator<Method> iter = map.keySet().iterator();
            while (iter.hasNext())
            {
               Method meth = iter.next();
               Class<?>[] parms = meth.getParameterTypes();
               int len = parms.length;
               if (len > 0 &&
                   Collection.class.isAssignableFrom(parms[len - 1]))
               {
                  iter.remove();
               }
            }
            
            map = Collections.unmodifiableMap(map);
            s2pCache.put(dmoIface, map);
         }
         
         return map;
      }
   }
   
   /**
    * Build a map of all methods in the given array, whose names begin with
    * any of the given prefixes.  The keys are property names, derived by
    * discarding a matching prefix from the method name, and lowercasing the
    * first letter of the remaining text.  The values are the methods.
    *
    * @param   methods
    *          Methods to be tested.
    * @param   prefix
    *          Array of prefix strings which are used to identify matching
    *          methods.
    *
    * @return  Map of DMO property names to matching methods.
    */
   static Map<String, Method> methodsByProperty(Method[] methods,
                                                String[] prefix)
   {
      // Always prefer indexed methods (specifically for setters, which have
      // other variants).
      Set<String> indexed = new HashSet<String>();
      
      Map<String, Method> map = new HashMap<String, Method>();
      int len = methods.length;
      for (int i = 0; i < len; i++)
      {
         Method next = methods[i];
         String name = next.getName();
         
         for (int j = 0; j < prefix.length; j++)
         {
            String pfx = prefix[j];
            if (name.startsWith(pfx))
            {
               String rootName = makePropertyName(name, pfx.length());
               
               // If we've already stored an indexed version, we don't want
               // to replace it.
               if (indexed.contains(rootName))
               {
                  continue;
               }
               
               map.put(rootName, next);
               
               // Check whether first parameter is an index (int).
               Class<?>[] argTypes = next.getParameterTypes();
               if (argTypes.length > 0 && Integer.TYPE.equals(argTypes[0]))
               {
                  indexed.add(rootName);
               }
               
               break;
            }
         }
      }
      
      if (map.isEmpty())
      {
         map = Collections.emptyMap();
      }
      
      return map;
   }
   
   /**
    * Build a map of all property names associated with the methods in the
    * given array, whose names begin with any of the given prefixes.  The
    * keys are the methods.  The values are the property names, derived by
    * discarding a matching prefix from the method name, and lowercasing the
    * first letter of the remaining text.
    *
    * @param   methods
    *          Methods to be tested.
    * @param   prefix
    *          Array of prefix strings which are used to identify matching
    *          methods.
    *
    * @return  Map of DMO methods to matching property names.
    */
   static Map<Method, String> propertiesByMethod(Method[] methods,
                                                 String[] prefix)
   {
      Map<Method, String> map = new HashMap<Method, String>();
      int len = methods.length;
      for (int i = 0; i < len; i++)
      {
         Method next = methods[i];
         String name = next.getName();
         
         for (int j = 0; j < prefix.length; j++)
         {
            String pfx = prefix[j];
            if (name.startsWith(pfx))
            {
               String rootName = makePropertyName(name, pfx.length());
               map.put(next, rootName);
               
               break;
            }
         }
      }
      
      return map;
   }
   
   /**
    * Get a set of all accessor (getter) methods for the given DMO interface.
    * <p>
    * Getter methods are all those declared by the interface, whose names
    * begin with <code>get</code> or <code>is</code>, excluding those which
    * return a collection.
    *
    * @param   dmoIface
    *          DMO interface to introspect.
    *
    * @return  Set of getter methods for the given interface.
    */
   static Set<Method> gettersForInterface(Class<?> dmoIface)
   {
      Set<Method> set = null;
      
      synchronized (i2gCache)
      {
         set = i2gCache.get(dmoIface);
         
         if (set == null)
         {
            Map<Method, String> g2p = propertiesByGetter(dmoIface);
            set = new HashSet<Method>(g2p.keySet());
            set = Collections.unmodifiableSet(set);
            i2gCache.put(dmoIface, set);
         }
      }
      
      return set;
   }
   
   /**
    * Get a list of methods which, when invoked, will trigger Hibernate to
    * initialize the lazily initialized persistent lists which are used to
    * manage extent fields for a DMO.  If a DMO has no extent fields, and
    * empty list is returned.
    * 
    * @param   dmoIface
    *          DMO interface to introspect.
    * @param   dmo
    *          DMO whose sizer methods (if any) will be invoked if a list is
    *          not already cached for this DMO type.
    * @param   dmoClass
    *          DMO implementation class.
    *
    * @return  A list of trigger methods or an empty list.
    * 
    * @throws  InvocationTargetException
    *          if an error occurs while using reflection to invoke the sizer
    *          methods of <code>dmo</code>.
    * @throws  IllegalAccessException
    *          if a sizer method of <code>dmo</code> cannot be accessed via
    *          reflection.
    */
   static List<Method> lazyExtentTriggersForInterface(Class<?> dmoIface,
                                                      Persistable dmo,
                                                      Class<?> dmoClass)
   throws InvocationTargetException,
          IllegalAccessException
   {
      List<Method> list = null;
      
      synchronized (i2ltCache)
      {
         list = i2ltCache.get(dmoIface);
         
         if (list == null)
         {
            Map<String, Method> getters = gettersByProperty(dmoIface);
            Map<Integer, Method> gettersBySize =
               new HashMap<Integer, Method>();
            Iterator<Map.Entry<String, Method>> iter =
               sizersByProperty(dmoIface).entrySet().iterator();
            String schema = DatabaseManager.getSchemaByClass(dmoClass);
            while (iter.hasNext())
            {
               Map.Entry<String, Method> next = iter.next();
               String property = next.getKey();
               if (!DatabaseManager.TEMP_TABLE_SCHEMA.equals(schema) &&
                   TableMapper.getLegacyFieldInfo(dmoClass, property) == null)
               {
                  continue;
               }
               Method sizer = next.getValue();
               Method getter = getters.get(property);
               integer i = (integer) sizer.invoke(dmo);
               Integer size = new Integer(i.intValue());
               Method bestSoFar = gettersBySize.get(size);
               if (bestSoFar == null)
               {
                  gettersBySize.put(size, getter);
               }
               else
               {
                  Class<?> bestTypeSoFar = bestSoFar.getReturnType();
                  if (integer.class.equals(bestTypeSoFar) ||
                      logical.class.equals(bestTypeSoFar))
                  {
                     // Can't improve on logical or integer;  their copy
                     // constructors are the fastest.
                     continue;
                  }
                  Class<?> type = getter.getReturnType();
                  if (decimal.class.equals(bestTypeSoFar) &&
                      !decimal.class.equals(type))
                  {
                     // Anything is an improvement over a decimal return type,
                     // which is the slowest to construct.
                     gettersBySize.put(size, getter);
                     continue;
                  }
               }
            }
            
            if (gettersBySize.isEmpty())
            {
               list = Collections.emptyList();
            }
            else
            {
               list = new ArrayList<Method>(gettersBySize.values());
               list = Collections.unmodifiableList(list);
            }
            
            i2ltCache.put(dmoIface, list);
         }
      }
      
      return list;
   }
   
   /**
    * Given a method name ending in a property name, extract the property
    * name by trimming off the prefix and decapitalizing the root name, using
    * <code>java.beans.Introspector.decapitalize()</code>.  Note that this
    * method will not decapitalize the first letter of the name if both the
    * first letter and the second letter (if any) are both uppercase.  This
    * quirk is necessary for compatibility with Hibernate's internal property
    * name conversions.
    *
    * @param   methodName
    *          Method name ending in the property name.
    * @param   index
    *          Index of first character of property name.
    *
    * @return  Property name.
    */
   static String makePropertyName(String methodName, int index)
   {
      String baseName = methodName.substring(index);
      String propName = Introspector.decapitalize(baseName);
      
      return propName.intern();
   }

   /**
    * Get property name for given getter/setter.
    *
    * @param    method
    *           Method.
    *
    * @return   Property name.
    */
   static String getPropertyName(Method method)
   {
      String methodName = method.getName();
      int index;
      if (methodName.startsWith("get") || methodName.startsWith("set"))
      {
         index = 3;
      }
      else if (methodName.startsWith("is"))
      {
         index = 2;
      }
      else
      {
         throw new UnsupportedOperationException(
               "Method getPropertyName is acceptable only for getters/setters");
      }
      return makePropertyName(methodName, index);
   }

}
